﻿using System;
using System.Diagnostics;

namespace TypeConstructors
{

    class WithoutConstructor
    {
        internal static Int32 Counter = 0;
    }
    class WithExplicitConstructor
    {
        internal static Int32 Counter = 0;
        static WithExplicitConstructor()
        {
        }
    }
    class Program
    {
        static void Main(string[] args)
        {

            Program p = new Program();
            ConsoleKeyInfo key;
            int iterations = 2000 * 1000 * 1000;
            do
            {


                p.Method1(iterations);
                Console.WriteLine();

                p.Method2(iterations);
                Console.WriteLine();
                Console.Write("Press Y to repeat, any other key to exit: ");
                key = Console.ReadKey();
                Console.WriteLine();
                Console.WriteLine();
            } while (key.Key == ConsoleKey.Y);
        }

        /// <summary>
        /// When this method starts to execute the JIT compiler adds the native code instructions to execute 
        /// the two static constructors within the scope of this method since they have not been run before.
        /// </summary>
        /// <param name="iterations"></param>
        void Method1(int iterations)
        {
            Console.WriteLine("Method 1 time lines follow");
            Stopwatch w = Stopwatch.StartNew();


            for (int i = 0; i < iterations; i++)
            {
                /* for this class the JIT compiler follows the Before-Field-Init semantics
                 * It is thus able to write the instructoins that verify whether the type constructor has been executed or not and execute it if needed
                 * outside the body of this loop resulting in this loop running faster.
                 * */
                WithoutConstructor.Counter = 0;
            }
            w.Stop();
            Console.WriteLine("Time taken by the loop for class with no defined type constructor: {0}", w.Elapsed.Seconds);

            w = Stopwatch.StartNew();

            for (int i = 0; i < iterations; i++)
            {
                /* Since the JIT compiler is follwing Precise semantics for WithExplicitConstructor class
                 * it emits the instructions to verify and execute the type constructor at this point.
                 * Each executing thread thus will run the verification logic with each iteration in addition to the 
                 * actual type constructor code being run by the first thread that gets here
                 * this results in performance hit since the verification code runs as many times as there are iterations
                 * */
                WithExplicitConstructor.Counter = 0;
            }
            w.Stop();
            Console.WriteLine("Time taken by the loop for class with the defined type constructor: {0}", w.Elapsed.Seconds);
        }

        /// <summary>
        /// When this method is called within the code, the JIT compiler knows that the two types used within the method
        /// have had their type constructors executed earlier on for this AppDomain. The JIT compiler thus does not emit the
        /// calls to verify and execute the two type constructors within the scope of this method.
        /// Thus the two loops within this method run with identical time lines.
        /// </summary>
        /// <param name="iterations"></param>
        void Method2(int iterations)
        {
            Console.WriteLine("Method 2 time lines follow");
            Stopwatch w = Stopwatch.StartNew(); ;

            for (int i = 0; i < iterations; i++)
            {
                WithoutConstructor.Counter = 0;
            }
            w.Stop();
            Console.WriteLine("Time taken by the loop for class with no defined type constructor: {0}", w.Elapsed.Seconds);


            w = Stopwatch.StartNew();

            for (int i = 0; i < iterations; i++)
            {
                WithExplicitConstructor.Counter = 0;
            }
            w.Stop();
            Console.WriteLine("Time taken by the loop for class with the defined type constructor: {0}", w.Elapsed.Seconds);
        }

    }


}

